@@ -72,6 +72,24 @@
  *
  * A same #GtkSourceCompletionProvider object can be used for several
  * #GtkSourceCompletion.
+ *
+ * # GtkSourceCompletion as GtkBuildable
+ *
+ * The GtkSourceCompletion implementation of the #GtkBuildable interface exposes
+ * the info window object (see gtk_source_completion_get_info_window()) with the
+ * internal-child "info_window".
+ *
+ * An example of a UI definition fragment with GtkSourceCompletion:
+ * |[
+ * <object class="GtkSourceCompletion">
+ *   <property name="select_on_show">False</property>
+ *   <child internal-child="info_window">
+ *     <object class="GtkSourceCompletionInfo">
+ *       <property name="border_width">6</property>
+ *     </object>
+ *   </child>
+ * </object>
+ * ]|
  */
 
 /* Idea to improve the code: use a composite widget template. This class is not
@@ -184,7 +202,12 @@ struct _GtkSourceCompletionPrivate
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
-G_DEFINE_TYPE_WITH_PRIVATE (GtkSourceCompletion, gtk_source_completion, G_TYPE_OBJECT)
+static void	gtk_source_completion_buildable_interface_init (GtkBuildableIface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (GtkSourceCompletion, gtk_source_completion, G_TYPE_OBJECT,
+			 G_ADD_PRIVATE (GtkSourceCompletion)
+			 G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
+						gtk_source_completion_buildable_interface_init))
 
 static void
 scroll_to_iter (GtkSourceCompletion *completion,
@@ -2027,6 +2050,8 @@ init_main_window (GtkSourceCompletion *completion,
 	completion->priv->selection_label = GTK_LABEL (gtk_builder_get_object (builder, "selection_label"));
 	completion->priv->bottom_bar = GTK_WIDGET (gtk_builder_get_object (builder, "bottom_bar"));
 
+	gtk_container_set_border_width (GTK_CONTAINER (completion->priv->main_window), 0);
+
 	gtk_window_set_attached_to (GTK_WINDOW (completion->priv->main_window),
 				    GTK_WIDGET (completion->priv->view));
 
@@ -2357,7 +2382,7 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 	 * <keycap>Home</keycap> and <keycap>End</keycap> keys are bound to the
 	 * normal behavior expected by those keys.
 	 *
-	 * When @step is equal to #GTK_SCROLL_PAGES, the page size is defined by
+	 * When @step is equal to %GTK_SCROLL_PAGES, the page size is defined by
 	 * the #GtkSourceCompletion:proposal-page-size property. It is used for
 	 * the <keycap>PageDown</keycap> and <keycap>PageUp</keycap> keys.
 	 *
@@ -2541,6 +2566,27 @@ gtk_source_completion_init (GtkSourceCompletion *completion)
 	completion->priv = gtk_source_completion_get_instance_private (completion);
 }
 
+static GObject *
+gtk_source_completion_buildable_get_internal_child (GtkBuildable *buildable,
+						    GtkBuilder   *builder,
+						    const gchar  *childname)
+{
+	GtkSourceCompletion *completion = GTK_SOURCE_COMPLETION (buildable);
+
+	if (g_strcmp0 (childname, "info_window") == 0)
+	{
+		return G_OBJECT (gtk_source_completion_get_info_window (completion));
+	}
+
+	return NULL;
+}
+
+static void
+gtk_source_completion_buildable_interface_init (GtkBuildableIface *iface)
+{
+	iface->get_internal_child = gtk_source_completion_buildable_get_internal_child;
+}
+
 void
 _gtk_source_completion_add_proposals (GtkSourceCompletion         *completion,
                                       GtkSourceCompletionContext  *context,
