@@ -394,49 +394,14 @@ calculate_gutter_size (GtkSourceGutter  *gutter,
 	return total_width;
 }
 
-static void
-window_invalidate_handler (GdkWindow      *window,
-			   cairo_region_t *region)
-{
-	cairo_rectangle_int_t rect;
-
-	/* Always invalidate the whole window.
-	 * When the text is modified in a GtkTextBuffer, GtkTextView tries to
-	 * redraw the smallest required region. But the information displayed in
-	 * the gutter may become invalid in a bigger region.
-	 * See https://bugzilla.gnome.org/show_bug.cgi?id=732418 for an example
-	 * where line numbers are not updated correctly when splitting a wrapped
-	 * line.
-	 * The performances should not be a big problem here. Correctness is
-	 * more important than performances.
-	 */
-
-	rect.x = 0;
-	rect.y = 0;
-	rect.width = gdk_window_get_width (window);
-	rect.height = gdk_window_get_height (window);
-
-	cairo_region_union_rectangle (region, &rect);
-}
-
 static void
 update_gutter_size (GtkSourceGutter *gutter)
 {
-	gint width;
-	GdkWindow *window;
-
-	width = calculate_gutter_size (gutter, NULL);
+	gint width = calculate_gutter_size (gutter, NULL);
 
 	gtk_text_view_set_border_window_size (GTK_TEXT_VIEW (gutter->priv->view),
 	                                      gutter->priv->window_type,
 	                                      width);
-
-	window = get_window (gutter);
-
-	if (window != NULL)
-	{
-		gdk_window_set_invalidate_handler (window, window_invalidate_handler);
-	}
 }
 
 static gboolean
@@ -957,6 +922,7 @@ on_view_draw (GtkSourceView   *view,
 	gboolean has_selection;
 	gint idx;
 	GtkStyleContext *style_context;
+	const gchar *class;
 	GdkRGBA fg_color;
 
 	window = get_window (gutter);
@@ -968,8 +934,6 @@ on_view_draw (GtkSourceView   *view,
 
 	gtk_cairo_transform_to_window (cr, GTK_WIDGET (view), window);
 
-	text_view = GTK_TEXT_VIEW (view);
-
 	if (!gdk_cairo_get_clip_rectangle (cr, &clip))
 	{
 		return FALSE;
@@ -977,6 +941,7 @@ on_view_draw (GtkSourceView   *view,
 
 	gutter->priv->is_drawing = TRUE;
 
+	text_view = GTK_TEXT_VIEW (view);
 	buffer = gtk_text_view_get_buffer (text_view);
 
 	y1 = clip.y;
@@ -1029,6 +994,30 @@ on_view_draw (GtkSourceView   *view,
 
 	style_context = gtk_widget_get_style_context (GTK_WIDGET (view));
 
+	gtk_style_context_save (style_context);
+
+	switch (gutter->priv->window_type)
+	{
+		case GTK_TEXT_WINDOW_TOP:
+			class = GTK_STYLE_CLASS_TOP;
+			break;
+		case GTK_TEXT_WINDOW_RIGHT:
+			class = GTK_STYLE_CLASS_RIGHT;
+			break;
+		case GTK_TEXT_WINDOW_BOTTOM:
+			class = GTK_STYLE_CLASS_BOTTOM;
+			break;
+		case GTK_TEXT_WINDOW_LEFT:
+			class = GTK_STYLE_CLASS_LEFT;
+			break;
+		default:
+			g_return_val_if_reached (FALSE);
+	}
+
+	/* Apply classes ourselves, since we are in connect_after and so they
+	 * are not set by gtk */
+	gtk_style_context_add_class (style_context, class);
+
 	gtk_style_context_get_color (style_context,
 	                             gtk_widget_get_state_flags (GTK_WIDGET (view)),
 	                             &fg_color);
@@ -1202,6 +1191,9 @@ on_view_draw (GtkSourceView   *view,
 		gtk_text_iter_forward_line (&start);
 	}
 
+	/* Allow to call queue_redraw() in ::end. */
+	gutter->priv->is_drawing = FALSE;
+
 	for (item = gutter->priv->renderers; item != NULL; item = g_list_next (item))
 	{
 		Renderer *renderer = item->data;
@@ -1217,7 +1209,7 @@ on_view_draw (GtkSourceView   *view,
 	g_array_free (heights, TRUE);
 	g_array_free (sizes, TRUE);
 
-	gutter->priv->is_drawing = FALSE;
+	gtk_style_context_restore (style_context);
 
 	return FALSE;
 }
