@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * gtksourceundomanager.c
+ * gtksourceundomanagerdefault.c
  * This file is part of GtkSourceView
  *
  * Copyright (C) 1998, 1999 Alex Roberts, Evan Lawrence
@@ -16,7 +16,7 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
-
+ *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
@@ -26,15 +26,14 @@
 #include <config.h>
 #endif
 
-#include <glib.h>
-#include <stdlib.h>
+#include "gtksourceundomanagerdefault.h"
+
 #include <string.h>
 
-#include "gtksourceundomanagerdefault.h"
 #include "gtksourceundomanager.h"
 #include "gtksourceview-i18n.h"
 
-#define DEFAULT_MAX_UNDO_LEVELS		-1
+#define DEFAULT_MAX_UNDO_LEVELS -1
 
 /*
  * The old code which used a GSList and g_slist_nth_element
@@ -47,9 +46,9 @@
  * structures or something.
  */
 
-typedef struct _GtkSourceUndoAction  			GtkSourceUndoAction;
-typedef struct _GtkSourceUndoInsertAction		GtkSourceUndoInsertAction;
-typedef struct _GtkSourceUndoDeleteAction		GtkSourceUndoDeleteAction;
+typedef struct _GtkSourceUndoAction		GtkSourceUndoAction;
+typedef struct _GtkSourceUndoInsertAction	GtkSourceUndoInsertAction;
+typedef struct _GtkSourceUndoDeleteAction	GtkSourceUndoDeleteAction;
 
 typedef enum
 {
@@ -57,25 +56,24 @@ typedef enum
 	GTK_SOURCE_UNDO_ACTION_DELETE
 } GtkSourceUndoActionType;
 
-/*
- * We use offsets instead of GtkTextIters because the last ones
- * require to much memory in this context without giving us any advantage.
+/* We use offsets instead of GtkTextMarks because the latter require too much
+ * memory in this context without giving us any advantage.
  */
 
 struct _GtkSourceUndoInsertAction
 {
-	gint   pos;
+	gint pos;
 	gchar *text;
-	gint   length;
-	gint   chars;
+	gint length;
+	gint chars;
 };
 
 struct _GtkSourceUndoDeleteAction
 {
-	gint      start;
-	gint      end;
-	gchar    *text;
-	gboolean  forward;
+	gint start;
+	gint end;
+	gchar *text;
+	guint forward : 1;
 };
 
 struct _GtkSourceUndoAction
@@ -84,8 +82,8 @@ struct _GtkSourceUndoAction
 
 	union
 	{
-		GtkSourceUndoInsertAction  insert;
-		GtkSourceUndoDeleteAction  delete;
+		GtkSourceUndoInsertAction insert;
+		GtkSourceUndoDeleteAction delete;
 	} action;
 
 	gint selection_insert;
@@ -102,23 +100,36 @@ struct _GtkSourceUndoAction
 	 * action of a group can be marked as modified.
 	 * There can be a single action marked as "modified" in the actions list.
 	 */
-	guint modified  : 1;
+	guint modified : 1;
 };
 
 enum
 {
-	INSERT_TEXT,
-	DELETE_RANGE,
-	BEGIN_USER_ACTION,
-	MODIFIED_CHANGED,
-	NUM_SIGNALS
+	PROP_0,
+	PROP_BUFFER,
+	PROP_MAX_UNDO_LEVELS
 };
 
 struct _GtkSourceUndoManagerDefaultPrivate
 {
+	/* Weak ref to the buffer. */
 	GtkTextBuffer *buffer;
 
+	/* Array of GtkSourceUndoAction's. The most recent action is at the end
+	 * of the array. But we can be in the middle of the array if there are
+	 * redo actions on the right.
+	 */
 	GPtrArray *actions;
+
+	/* Index of the next redo action. Attention, the index is counted from
+	 * the end of the array:
+	 * - next_redo of -1: there is no redo actions.
+	 * - next_redo of 0: there is one redo action, the last item in
+	 *   'actions'.
+	 * - next_redo of actions->len - 1: there is no undo actions, the next
+	 *   redo action is the first item in 'actions'.
+	 * - the next undo action is located at next_redo + 1.
+	 */
 	gint next_redo;
 
 	gint actions_in_current_group;
@@ -134,49 +145,12 @@ struct _GtkSourceUndoManagerDefaultPrivate
 	 */
 	guint modified_undoing_group : 1;
 
-	/* Pointer to the action (in the action list) marked as "modified".
-	 * It is NULL when no action is marked as "modified". */
+	/* Pointer to the action (in 'actions') marked as "modified".
+	 * It is NULL when no action is marked as "modified".
+	 */
 	GtkSourceUndoAction *modified_action;
-
-	guint buffer_signals[NUM_SIGNALS];
 };
 
-/* Properties */
-enum
-{
-	PROP_0,
-	PROP_BUFFER,
-	PROP_MAX_UNDO_LEVELS
-};
-
-static void insert_text_handler       (GtkTextBuffer             *buffer,
-                                       GtkTextIter               *pos,
-                                       const gchar               *text,
-                                       gint                       length,
-                                       GtkSourceUndoManagerDefault      *um);
-
-static void delete_range_handler      (GtkTextBuffer             *buffer,
-                                       GtkTextIter               *start,
-                                       GtkTextIter               *end,
-                                       GtkSourceUndoManagerDefault      *um);
-
-static void begin_user_action_handler (GtkTextBuffer             *buffer,
-                                       GtkSourceUndoManagerDefault      *um);
-
-static void modified_changed_handler  (GtkTextBuffer             *buffer,
-                                       GtkSourceUndoManagerDefault      *um);
-
-static void free_action_list          (GtkSourceUndoManagerDefault      *um);
-
-static void add_action                (GtkSourceUndoManagerDefault      *um,
-                                       const GtkSourceUndoAction *undo_action);
-static void free_first_n_actions      (GtkSourceUndoManagerDefault      *um,
-                                       gint                       n);
-static void check_list_size           (GtkSourceUndoManagerDefault      *um);
-
-static gboolean merge_action          (GtkSourceUndoManagerDefault      *um,
-                                       const GtkSourceUndoAction *undo_action);
-
 static void gtk_source_undo_manager_iface_init (GtkSourceUndoManagerIface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (GtkSourceUndoManagerDefault, gtk_source_undo_manager_default, G_TYPE_OBJECT,
@@ -185,16 +159,336 @@ G_DEFINE_TYPE_WITH_CODE (GtkSourceUndoManagerDefault, gtk_source_undo_manager_de
                                                 gtk_source_undo_manager_iface_init))
 
 static void
-gtk_source_undo_manager_default_finalize (GObject *object)
+gtk_source_undo_action_free (GtkSourceUndoAction *action)
 {
-	GtkSourceUndoManagerDefault *manager;
+	if (action == NULL)
+	{
+		return;
+	}
 
-	manager = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
+	if (action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
+	{
+		g_free (action->action.insert.text);
+	}
+	else if (action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
+	{
+		g_free (action->action.delete.text);
+	}
+	else
+	{
+		g_free (action);
+		g_return_if_reached ();
+	}
 
-	free_action_list (manager);
-	g_ptr_array_free (manager->priv->actions, TRUE);
+	g_free (action);
+}
 
-	G_OBJECT_CLASS (gtk_source_undo_manager_default_parent_class)->finalize (object);
+static void
+free_action_list (GtkSourceUndoManagerDefault *um)
+{
+	gint i;
+
+	for (i = (gint)um->priv->actions->len - 1; i >= 0; i--)
+	{
+		GtkSourceUndoAction *action = um->priv->actions->pdata[i];
+
+		if (action->order_in_group == 1)
+			--um->priv->num_of_groups;
+
+		if (action->modified)
+			um->priv->modified_action = NULL;
+
+		gtk_source_undo_action_free (action);
+	}
+
+	/* Some arbitrary limit, to avoid wasting space */
+	if (um->priv->actions->len > 2048)
+	{
+		g_ptr_array_free (um->priv->actions, TRUE);
+		um->priv->actions = g_ptr_array_new ();
+	}
+	else
+	{
+		g_ptr_array_set_size (um->priv->actions, 0);
+	}
+}
+
+static GtkSourceUndoAction *
+action_list_nth_data (GPtrArray *array,
+                      gint       n)
+{
+	if (n < 0 || n >= (gint)array->len)
+		return NULL;
+	else
+		return array->pdata[array->len - 1 - n];
+}
+
+static void
+action_list_prepend (GPtrArray           *array,
+                     GtkSourceUndoAction *action)
+{
+	g_ptr_array_add (array, action);
+}
+
+static GtkSourceUndoAction *
+action_list_last_data (GPtrArray *array)
+{
+	if (array->len != 0)
+		return array->pdata[0];
+	else
+		return NULL;
+}
+
+static void
+action_list_delete_last (GPtrArray *array)
+{
+	if (array->len != 0)
+	{
+		memmove (&array->pdata[0], &array->pdata[1], (array->len - 1)*sizeof (gpointer));
+		g_ptr_array_set_size (array, array->len - 1);
+	}
+}
+
+static void
+free_first_n_actions (GtkSourceUndoManagerDefault *um,
+                      gint                         n)
+{
+	gint i;
+
+	if (um->priv->actions->len == 0)
+		return;
+
+	for (i = 0; i < n; i++)
+	{
+		GtkSourceUndoAction *action = um->priv->actions->pdata[um->priv->actions->len - 1];
+
+		if (action->order_in_group == 1)
+			--um->priv->num_of_groups;
+
+		if (action->modified)
+			um->priv->modified_action = NULL;
+
+		gtk_source_undo_action_free (action);
+
+		g_ptr_array_set_size (um->priv->actions, um->priv->actions->len - 1);
+
+		if (um->priv->actions->len == 0)
+			return;
+	}
+}
+
+static void
+check_list_size (GtkSourceUndoManagerDefault *um)
+{
+	gint undo_levels;
+
+	undo_levels = um->priv->max_undo_levels;
+
+	if (undo_levels < 1)
+		return;
+
+	if (um->priv->num_of_groups > undo_levels)
+	{
+		GtkSourceUndoAction *undo_action;
+
+		undo_action = action_list_last_data (um->priv->actions);
+
+		do
+		{
+			if (undo_action->order_in_group == 1)
+				--um->priv->num_of_groups;
+
+			if (undo_action->modified)
+				um->priv->modified_action = NULL;
+
+			gtk_source_undo_action_free (undo_action);
+
+			action_list_delete_last (um->priv->actions);
+
+			undo_action = action_list_last_data (um->priv->actions);
+			g_return_if_fail (undo_action != NULL);
+
+		} while ((undo_action->order_in_group > 1) ||
+			 (um->priv->num_of_groups > undo_levels));
+	}
+}
+
+/*
+ * merge_action:
+ * @um: a #GtkSourceUndoManagerDefault.
+ * @undo_action: a #GtkSourceUndoAction.
+ *
+ * This function tries to merge the undo action at the top of
+ * the stack with a new undo action. So when we undo for example
+ * typing, we can undo the whole word and not each letter by itself.
+ *
+ * Return Value: %TRUE is merge was sucessful, %FALSE otherwise.
+ */
+static gboolean
+merge_action (GtkSourceUndoManagerDefault *um,
+              const GtkSourceUndoAction   *undo_action)
+{
+	GtkSourceUndoAction *last_action;
+
+	g_return_val_if_fail (GTK_SOURCE_IS_UNDO_MANAGER_DEFAULT (um), FALSE);
+	g_return_val_if_fail (um->priv != NULL, FALSE);
+
+	if (um->priv->actions->len == 0)
+		return FALSE;
+
+	last_action = action_list_nth_data (um->priv->actions, 0);
+
+	if (!last_action->mergeable)
+		return FALSE;
+
+	if ((!undo_action->mergeable) ||
+	    (undo_action->action_type != last_action->action_type))
+	{
+		last_action->mergeable = FALSE;
+		return FALSE;
+	}
+
+	if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
+	{
+		if ((last_action->action.delete.forward != undo_action->action.delete.forward) ||
+		    ((last_action->action.delete.start != undo_action->action.delete.start) &&
+		     (last_action->action.delete.start != undo_action->action.delete.end)))
+		{
+			last_action->mergeable = FALSE;
+			return FALSE;
+		}
+
+		if (last_action->action.delete.start == undo_action->action.delete.start)
+		{
+			gchar *str;
+
+#define L  (last_action->action.delete.end - last_action->action.delete.start - 1)
+#define g_utf8_get_char_at(p,i) g_utf8_get_char(g_utf8_offset_to_pointer((p),(i)))
+
+			/* Deleted with the delete key */
+			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
+			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
+                            ((g_utf8_get_char_at (last_action->action.delete.text, L) == ' ') ||
+			     (g_utf8_get_char_at (last_action->action.delete.text, L)  == '\t')))
+			{
+				last_action->mergeable = FALSE;
+				return FALSE;
+			}
+
+			str = g_strdup_printf ("%s%s", last_action->action.delete.text,
+				undo_action->action.delete.text);
+
+			g_free (last_action->action.delete.text);
+			last_action->action.delete.end += (undo_action->action.delete.end -
+							   undo_action->action.delete.start);
+			last_action->action.delete.text = str;
+		}
+		else
+		{
+			gchar *str;
+
+			/* Deleted with the backspace key */
+			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
+			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
+                            ((g_utf8_get_char (last_action->action.delete.text) == ' ') ||
+			     (g_utf8_get_char (last_action->action.delete.text) == '\t')))
+			{
+				last_action->mergeable = FALSE;
+				return FALSE;
+			}
+
+			str = g_strdup_printf ("%s%s", undo_action->action.delete.text,
+				last_action->action.delete.text);
+
+			g_free (last_action->action.delete.text);
+			last_action->action.delete.start = undo_action->action.delete.start;
+			last_action->action.delete.text = str;
+		}
+	}
+	else if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
+	{
+		gchar* str;
+
+#define I (last_action->action.insert.chars - 1)
+
+		if ((undo_action->action.insert.pos !=
+		     (last_action->action.insert.pos + last_action->action.insert.chars)) ||
+		    ((g_utf8_get_char (undo_action->action.insert.text) != ' ') &&
+		      (g_utf8_get_char (undo_action->action.insert.text) != '\t') &&
+		     ((g_utf8_get_char_at (last_action->action.insert.text, I) == ' ') ||
+		      (g_utf8_get_char_at (last_action->action.insert.text, I) == '\t')))
+		   )
+		{
+			last_action->mergeable = FALSE;
+			return FALSE;
+		}
+
+		str = g_strdup_printf ("%s%s", last_action->action.insert.text,
+				undo_action->action.insert.text);
+
+		g_free (last_action->action.insert.text);
+		last_action->action.insert.length += undo_action->action.insert.length;
+		last_action->action.insert.text = str;
+		last_action->action.insert.chars += undo_action->action.insert.chars;
+
+	}
+	else
+		/* Unknown action inside undo merge encountered */
+		g_return_val_if_reached (TRUE);
+
+	return TRUE;
+}
+
+static void
+add_action (GtkSourceUndoManagerDefault *um,
+            const GtkSourceUndoAction   *undo_action)
+{
+	GtkSourceUndoAction* action;
+
+	if (um->priv->next_redo >= 0)
+	{
+		free_first_n_actions (um, um->priv->next_redo + 1);
+	}
+
+	um->priv->next_redo = -1;
+
+	if (!merge_action (um, undo_action))
+	{
+		action = g_new (GtkSourceUndoAction, 1);
+		*action = *undo_action;
+
+		if (action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
+			action->action.insert.text = g_strndup (undo_action->action.insert.text, undo_action->action.insert.length);
+		else if (action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
+			action->action.delete.text = g_strdup (undo_action->action.delete.text);
+		else
+		{
+			g_free (action);
+			g_return_if_reached ();
+		}
+
+		++um->priv->actions_in_current_group;
+		action->order_in_group = um->priv->actions_in_current_group;
+
+		if (action->order_in_group == 1)
+			++um->priv->num_of_groups;
+
+		action_list_prepend (um->priv->actions, action);
+	}
+
+	check_list_size (um);
+
+	if (!um->priv->can_undo)
+	{
+		um->priv->can_undo = TRUE;
+		gtk_source_undo_manager_can_undo_changed (GTK_SOURCE_UNDO_MANAGER (um));
+	}
+
+	if (um->priv->can_redo)
+	{
+		um->priv->can_redo = FALSE;
+		gtk_source_undo_manager_can_redo_changed (GTK_SOURCE_UNDO_MANAGER (um));
+	}
 }
 
 static void
@@ -217,73 +511,6 @@ clear_undo (GtkSourceUndoManagerDefault *manager)
 	}
 }
 
-static void
-buffer_notify (GtkSourceUndoManagerDefault *manager,
-               gpointer                     where_the_object_was)
-{
-	manager->priv->buffer = NULL;
-}
-
-static void
-set_buffer (GtkSourceUndoManagerDefault *manager,
-            GtkTextBuffer               *buffer)
-{
-	if (buffer == manager->priv->buffer)
-	{
-		return;
-	}
-
-	clear_undo (manager);
-
-	if (manager->priv->buffer != NULL)
-	{
-		gint i;
-
-		for (i = 0; i < NUM_SIGNALS; ++i)
-		{
-			g_signal_handler_disconnect (manager->priv->buffer,
-			                             manager->priv->buffer_signals[i]);
-		}
-
-		g_object_weak_unref (G_OBJECT (manager->priv->buffer),
-		                     (GWeakNotify)buffer_notify,
-		                     manager);
-		manager->priv->buffer = NULL;
-	}
-
-	if (buffer != NULL)
-	{
-		manager->priv->buffer = buffer;
-		g_object_weak_ref (G_OBJECT (buffer),
-		                   (GWeakNotify)buffer_notify,
-		                   manager);
-
-		manager->priv->buffer_signals[INSERT_TEXT] =
-			g_signal_connect (buffer,
-			                  "insert-text",
-			                  G_CALLBACK (insert_text_handler),
-			                  manager);
-
-		manager->priv->buffer_signals[DELETE_RANGE] =
-			g_signal_connect (buffer,
-			                  "delete-range",
-			                  G_CALLBACK (delete_range_handler),
-			                  manager);
-
-		manager->priv->buffer_signals[BEGIN_USER_ACTION] =
-			g_signal_connect (buffer,
-			                  "begin-user-action",
-			                  G_CALLBACK (begin_user_action_handler),
-			                  manager);
-
-		manager->priv->buffer_signals[MODIFIED_CHANGED] =
-			g_signal_connect (buffer,
-			                  "modified-changed",
-			                  G_CALLBACK (modified_changed_handler),
-			                  manager);
-	}
-}
-
 static void
 set_max_undo_levels (GtkSourceUndoManagerDefault *manager,
                      gint                         max_undo_levels)
@@ -325,105 +552,6 @@ set_max_undo_levels (GtkSourceUndoManagerDefault *manager,
 	}
 }
 
-static void
-gtk_source_undo_manager_default_dispose (GObject *object)
-{
-	GtkSourceUndoManagerDefault *manager;
-
-	manager = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
-
-	if (manager->priv->buffer != NULL)
-	{
-		/* Clear the buffer */
-		set_buffer (manager, NULL);
-	}
-
-	G_OBJECT_CLASS (gtk_source_undo_manager_default_parent_class)->dispose (object);
-}
-
-static void
-gtk_source_undo_manager_default_set_property (GObject      *object,
-                                              guint         prop_id,
-                                              const GValue *value,
-                                              GParamSpec   *pspec)
-{
-	GtkSourceUndoManagerDefault *self = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
-
-	switch (prop_id)
-	{
-		case PROP_BUFFER:
-			set_buffer (self, g_value_get_object (value));
-		break;
-		case PROP_MAX_UNDO_LEVELS:
-			gtk_source_undo_manager_default_set_max_undo_levels (self,
-			                                                     g_value_get_int (value));
-		break;
-		default:
-			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-		break;
-	}
-}
-
-static void
-gtk_source_undo_manager_default_get_property (GObject    *object,
-                                              guint       prop_id,
-                                              GValue     *value,
-                                              GParamSpec *pspec)
-{
-	GtkSourceUndoManagerDefault *self = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
-
-	switch (prop_id)
-	{
-		case PROP_BUFFER:
-			g_value_set_object (value, self->priv->buffer);
-		break;
-		case PROP_MAX_UNDO_LEVELS:
-			g_value_set_int (value, self->priv->max_undo_levels);
-		break;
-		default:
-			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-		break;
-	}
-}
-
-static void
-gtk_source_undo_manager_default_class_init (GtkSourceUndoManagerDefaultClass *klass)
-{
-	GObjectClass *object_class = G_OBJECT_CLASS (klass);
-
-	object_class->finalize = gtk_source_undo_manager_default_finalize;
-	object_class->dispose = gtk_source_undo_manager_default_dispose;
-
-	object_class->set_property = gtk_source_undo_manager_default_set_property;
-	object_class->get_property = gtk_source_undo_manager_default_get_property;
-
-	g_object_class_install_property (object_class,
-	                                 PROP_BUFFER,
-	                                 g_param_spec_object ("buffer",
-	                                                      _("Buffer"),
-	                                                      _("The text buffer to add undo support on"),
-	                                                      GTK_TYPE_TEXT_BUFFER,
-	                                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
-
-	g_object_class_install_property (object_class,
-	                                 PROP_MAX_UNDO_LEVELS,
-	                                 g_param_spec_int ("max-undo-levels",
-	                                                   _("Maximum Undo Levels"),
-	                                                   _("Number of undo levels for "
-							     "the buffer"),
-	                                                   -1,
-	                                                   G_MAXINT,
-	                                                   DEFAULT_MAX_UNDO_LEVELS,
-	                                                   G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
-}
-
-static void
-gtk_source_undo_manager_default_init (GtkSourceUndoManagerDefault *um)
-{
-	um->priv = gtk_source_undo_manager_default_get_instance_private (um);
-	um->priv->actions = g_ptr_array_new ();
-}
-
 static void
 end_not_undoable_action_internal (GtkSourceUndoManagerDefault *manager)
 {
@@ -543,42 +671,6 @@ get_chars (GtkTextBuffer *buffer,
 	return gtk_text_buffer_get_slice (buffer, &start_iter, &end_iter, TRUE);
 }
 
-static GtkSourceUndoAction *
-action_list_nth_data (GPtrArray *array,
-                      gint       n)
-{
-	if (n < 0 || n >= (gint)array->len)
-		return NULL;
-	else
-		return array->pdata[array->len - 1 - n];
-}
-
-static void
-action_list_prepend (GPtrArray           *array,
-                     GtkSourceUndoAction *action)
-{
-	g_ptr_array_add (array, action);
-}
-
-static GtkSourceUndoAction *
-action_list_last_data (GPtrArray *array)
-{
-	if (array->len != 0)
-		return array->pdata[0];
-	else
-		return NULL;
-}
-
-static void
-action_list_delete_last (GPtrArray *array)
-{
-	if (array->len != 0)
-	{
-		memmove (&array->pdata[0], &array->pdata[1], (array->len - 1)*sizeof (gpointer));
-		g_ptr_array_set_size (array, array->len - 1);
-	}
-}
-
 static void
 selection_bounds_offsets (GtkTextBuffer *buffer,
                           gint          *insert_offset,
@@ -831,66 +923,11 @@ gtk_source_undo_manager_redo_impl (GtkSourceUndoManager *manager)
 }
 
 static void
-gtk_source_undo_action_free (GtkSourceUndoAction *action)
-{
-	if (action == NULL)
-	{
-		return;
-	}
-
-	if (action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
-	{
-		g_free (action->action.insert.text);
-	}
-	else if (action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-	{
-		g_free (action->action.delete.text);
-	}
-	else
-	{
-		g_free (action);
-		g_return_if_reached ();
-	}
-
-	g_free (action);
-}
-
-static void
-free_action_list (GtkSourceUndoManagerDefault *um)
-{
-	gint i;
-
-	for (i = (gint)um->priv->actions->len - 1; i >= 0; i--)
-	{
-		GtkSourceUndoAction *action = um->priv->actions->pdata[i];
-
-		if (action->order_in_group == 1)
-			--um->priv->num_of_groups;
-
-		if (action->modified)
-			um->priv->modified_action = NULL;
-
-		gtk_source_undo_action_free (action);
-	}
-
-	/* Some arbitrary limit, to avoid wasting space */
-	if (um->priv->actions->len > 2048)
-	{
-		g_ptr_array_free (um->priv->actions, TRUE);
-		um->priv->actions = g_ptr_array_new ();
-	}
-	else
-	{
-		g_ptr_array_set_size (um->priv->actions, 0);
-	}
-}
-
-static void
-insert_text_handler (GtkTextBuffer               *buffer,
-                     GtkTextIter                 *pos,
-                     const gchar                 *text,
-                     gint                         length,
-                     GtkSourceUndoManagerDefault *manager)
+insert_text_cb (GtkTextBuffer               *buffer,
+		GtkTextIter                 *pos,
+		const gchar                 *text,
+		gint                         length,
+		GtkSourceUndoManagerDefault *manager)
 {
 	GtkSourceUndoAction undo_action;
 
@@ -918,10 +955,10 @@ insert_text_handler (GtkTextBuffer               *buffer,
 }
 
 static void
-delete_range_handler (GtkTextBuffer               *buffer,
-                      GtkTextIter                 *start,
-                      GtkTextIter                 *end,
-                      GtkSourceUndoManagerDefault *um)
+delete_range_cb (GtkTextBuffer               *buffer,
+		 GtkTextIter                 *start,
+		 GtkTextIter                 *end,
+		 GtkSourceUndoManagerDefault *um)
 {
 	GtkSourceUndoAction undo_action;
 	GtkTextIter insert_iter;
@@ -966,8 +1003,8 @@ delete_range_handler (GtkTextBuffer               *buffer,
 }
 
 static void
-begin_user_action_handler (GtkTextBuffer               *buffer,
-                           GtkSourceUndoManagerDefault *um)
+begin_user_action_cb (GtkTextBuffer               *buffer,
+		      GtkSourceUndoManagerDefault *um)
 {
 	if (um->priv->running_not_undoable_actions > 0)
 		return;
@@ -976,250 +1013,8 @@ begin_user_action_handler (GtkTextBuffer               *buffer,
 }
 
 static void
-add_action (GtkSourceUndoManagerDefault *um,
-            const GtkSourceUndoAction   *undo_action)
-{
-	GtkSourceUndoAction* action;
-
-	if (um->priv->next_redo >= 0)
-	{
-		free_first_n_actions (um, um->priv->next_redo + 1);
-	}
-
-	um->priv->next_redo = -1;
-
-	if (!merge_action (um, undo_action))
-	{
-		action = g_new (GtkSourceUndoAction, 1);
-		*action = *undo_action;
-
-		if (action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
-			action->action.insert.text = g_strndup (undo_action->action.insert.text, undo_action->action.insert.length);
-		else if (action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-			action->action.delete.text = g_strdup (undo_action->action.delete.text);
-		else
-		{
-			g_free (action);
-			g_return_if_reached ();
-		}
-
-		++um->priv->actions_in_current_group;
-		action->order_in_group = um->priv->actions_in_current_group;
-
-		if (action->order_in_group == 1)
-			++um->priv->num_of_groups;
-
-		action_list_prepend (um->priv->actions, action);
-	}
-
-	check_list_size (um);
-
-	if (!um->priv->can_undo)
-	{
-		um->priv->can_undo = TRUE;
-		gtk_source_undo_manager_can_undo_changed (GTK_SOURCE_UNDO_MANAGER (um));
-	}
-
-	if (um->priv->can_redo)
-	{
-		um->priv->can_redo = FALSE;
-		gtk_source_undo_manager_can_redo_changed (GTK_SOURCE_UNDO_MANAGER (um));
-	}
-}
-
-static void
-free_first_n_actions (GtkSourceUndoManagerDefault *um,
-                      gint                         n)
-{
-	gint i;
-
-	if (um->priv->actions->len == 0)
-		return;
-
-	for (i = 0; i < n; i++)
-	{
-		GtkSourceUndoAction *action = um->priv->actions->pdata[um->priv->actions->len - 1];
-
-		if (action->order_in_group == 1)
-			--um->priv->num_of_groups;
-
-		if (action->modified)
-			um->priv->modified_action = NULL;
-
-		gtk_source_undo_action_free (action);
-
-		g_ptr_array_set_size (um->priv->actions, um->priv->actions->len - 1);
-
-		if (um->priv->actions->len == 0)
-			return;
-	}
-}
-
-static void
-check_list_size (GtkSourceUndoManagerDefault *um)
-{
-	gint undo_levels;
-
-	undo_levels = um->priv->max_undo_levels;
-
-	if (undo_levels < 1)
-		return;
-
-	if (um->priv->num_of_groups > undo_levels)
-	{
-		GtkSourceUndoAction *undo_action;
-
-		undo_action = action_list_last_data (um->priv->actions);
-
-		do
-		{
-			if (undo_action->order_in_group == 1)
-				--um->priv->num_of_groups;
-
-			if (undo_action->modified)
-				um->priv->modified_action = NULL;
-
-			gtk_source_undo_action_free (undo_action);
-
-			action_list_delete_last (um->priv->actions);
-
-			undo_action = action_list_last_data (um->priv->actions);
-			g_return_if_fail (undo_action != NULL);
-
-		} while ((undo_action->order_in_group > 1) ||
-			 (um->priv->num_of_groups > undo_levels));
-	}
-}
-
-/**
- * gtk_source_undo_manager_default_merge_action:
- * @um: a #GtkSourceUndoManagerDefault.
- * @undo_action: a #GtkSourceUndoAction.
- *
- * This function tries to merge the undo action at the top of
- * the stack with a new undo action. So when we undo for example
- * typing, we can undo the whole word and not each letter by itself.
- *
- * Return Value: %TRUE is merge was sucessful, %FALSE otherwise.
- **/
-static gboolean
-merge_action (GtkSourceUndoManagerDefault *um,
-              const GtkSourceUndoAction   *undo_action)
-{
-	GtkSourceUndoAction *last_action;
-
-	g_return_val_if_fail (GTK_SOURCE_IS_UNDO_MANAGER_DEFAULT (um), FALSE);
-	g_return_val_if_fail (um->priv != NULL, FALSE);
-
-	if (um->priv->actions->len == 0)
-		return FALSE;
-
-	last_action = action_list_nth_data (um->priv->actions, 0);
-
-	if (!last_action->mergeable)
-		return FALSE;
-
-	if ((!undo_action->mergeable) ||
-	    (undo_action->action_type != last_action->action_type))
-	{
-		last_action->mergeable = FALSE;
-		return FALSE;
-	}
-
-	if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_DELETE)
-	{
-		if ((last_action->action.delete.forward != undo_action->action.delete.forward) ||
-		    ((last_action->action.delete.start != undo_action->action.delete.start) &&
-		     (last_action->action.delete.start != undo_action->action.delete.end)))
-		{
-			last_action->mergeable = FALSE;
-			return FALSE;
-		}
-
-		if (last_action->action.delete.start == undo_action->action.delete.start)
-		{
-			gchar *str;
-
-#define L  (last_action->action.delete.end - last_action->action.delete.start - 1)
-#define g_utf8_get_char_at(p,i) g_utf8_get_char(g_utf8_offset_to_pointer((p),(i)))
-
-			/* Deleted with the delete key */
-			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
-			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
-                            ((g_utf8_get_char_at (last_action->action.delete.text, L) == ' ') ||
-			     (g_utf8_get_char_at (last_action->action.delete.text, L)  == '\t')))
-			{
-				last_action->mergeable = FALSE;
-				return FALSE;
-			}
-
-			str = g_strdup_printf ("%s%s", last_action->action.delete.text,
-				undo_action->action.delete.text);
-
-			g_free (last_action->action.delete.text);
-			last_action->action.delete.end += (undo_action->action.delete.end -
-							   undo_action->action.delete.start);
-			last_action->action.delete.text = str;
-		}
-		else
-		{
-			gchar *str;
-
-			/* Deleted with the backspace key */
-			if ((g_utf8_get_char (undo_action->action.delete.text) != ' ') &&
-			    (g_utf8_get_char (undo_action->action.delete.text) != '\t') &&
-                            ((g_utf8_get_char (last_action->action.delete.text) == ' ') ||
-			     (g_utf8_get_char (last_action->action.delete.text) == '\t')))
-			{
-				last_action->mergeable = FALSE;
-				return FALSE;
-			}
-
-			str = g_strdup_printf ("%s%s", undo_action->action.delete.text,
-				last_action->action.delete.text);
-
-			g_free (last_action->action.delete.text);
-			last_action->action.delete.start = undo_action->action.delete.start;
-			last_action->action.delete.text = str;
-		}
-	}
-	else if (undo_action->action_type == GTK_SOURCE_UNDO_ACTION_INSERT)
-	{
-		gchar* str;
-
-#define I (last_action->action.insert.chars - 1)
-
-		if ((undo_action->action.insert.pos !=
-		     	(last_action->action.insert.pos + last_action->action.insert.chars)) ||
-		    ((g_utf8_get_char (undo_action->action.insert.text) != ' ') &&
-		      (g_utf8_get_char (undo_action->action.insert.text) != '\t') &&
-		     ((g_utf8_get_char_at (last_action->action.insert.text, I) == ' ') ||
-		      (g_utf8_get_char_at (last_action->action.insert.text, I) == '\t')))
-		   )
-		{
-			last_action->mergeable = FALSE;
-			return FALSE;
-		}
-
-		str = g_strdup_printf ("%s%s", last_action->action.insert.text,
-				undo_action->action.insert.text);
-
-		g_free (last_action->action.insert.text);
-		last_action->action.insert.length += undo_action->action.insert.length;
-		last_action->action.insert.text = str;
-		last_action->action.insert.chars += undo_action->action.insert.chars;
-
-	}
-	else
-		/* Unknown action inside undo merge encountered */
-		g_return_val_if_reached (TRUE);
-
-	return TRUE;
-}
-
-static void
-modified_changed_handler (GtkTextBuffer               *buffer,
-                          GtkSourceUndoManagerDefault *manager)
+modified_changed_cb (GtkTextBuffer               *buffer,
+		     GtkSourceUndoManagerDefault *manager)
 {
 	GtkSourceUndoAction *action;
 	gint idx;
@@ -1270,6 +1065,162 @@ modified_changed_handler (GtkTextBuffer               *buffer,
 	manager->priv->modified_action = action;
 }
 
+static void
+set_buffer (GtkSourceUndoManagerDefault *manager,
+            GtkTextBuffer               *buffer)
+{
+	g_assert (manager->priv->buffer == NULL);
+
+	if (buffer == NULL)
+	{
+		return;
+	}
+
+	manager->priv->buffer = buffer;
+
+	g_object_add_weak_pointer (G_OBJECT (buffer),
+				   (gpointer *)&manager->priv->buffer);
+
+	g_signal_connect_object (buffer,
+				 "insert-text",
+				 G_CALLBACK (insert_text_cb),
+				 manager,
+				 0);
+
+	g_signal_connect_object (buffer,
+				 "delete-range",
+				 G_CALLBACK (delete_range_cb),
+				 manager,
+				 0);
+
+	g_signal_connect_object (buffer,
+				 "begin-user-action",
+				 G_CALLBACK (begin_user_action_cb),
+				 manager,
+				 0);
+
+	g_signal_connect_object (buffer,
+				 "modified-changed",
+				 G_CALLBACK (modified_changed_cb),
+				 manager,
+				 0);
+}
+
+static void
+gtk_source_undo_manager_default_dispose (GObject *object)
+{
+	GtkSourceUndoManagerDefault *manager = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
+
+	if (manager->priv->buffer != NULL)
+	{
+		clear_undo (manager);
+
+		g_object_remove_weak_pointer (G_OBJECT (manager->priv->buffer),
+					      (gpointer *)&manager->priv->buffer);
+
+		manager->priv->buffer = NULL;
+	}
+
+	G_OBJECT_CLASS (gtk_source_undo_manager_default_parent_class)->dispose (object);
+}
+
+static void
+gtk_source_undo_manager_default_finalize (GObject *object)
+{
+	GtkSourceUndoManagerDefault *manager = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
+
+	free_action_list (manager);
+	g_ptr_array_free (manager->priv->actions, TRUE);
+
+	G_OBJECT_CLASS (gtk_source_undo_manager_default_parent_class)->finalize (object);
+}
+
+static void
+gtk_source_undo_manager_default_set_property (GObject      *object,
+                                              guint         prop_id,
+                                              const GValue *value,
+                                              GParamSpec   *pspec)
+{
+	GtkSourceUndoManagerDefault *manager = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
+
+	switch (prop_id)
+	{
+		case PROP_BUFFER:
+			set_buffer (manager, g_value_get_object (value));
+			break;
+
+		case PROP_MAX_UNDO_LEVELS:
+			gtk_source_undo_manager_default_set_max_undo_levels (manager, g_value_get_int (value));
+			break;
+
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+			break;
+	}
+}
+
+static void
+gtk_source_undo_manager_default_get_property (GObject    *object,
+                                              guint       prop_id,
+                                              GValue     *value,
+                                              GParamSpec *pspec)
+{
+	GtkSourceUndoManagerDefault *manager = GTK_SOURCE_UNDO_MANAGER_DEFAULT (object);
+
+	switch (prop_id)
+	{
+		case PROP_BUFFER:
+			g_value_set_object (value, manager->priv->buffer);
+			break;
+
+		case PROP_MAX_UNDO_LEVELS:
+			g_value_set_int (value, manager->priv->max_undo_levels);
+			break;
+
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+			break;
+	}
+}
+
+static void
+gtk_source_undo_manager_default_class_init (GtkSourceUndoManagerDefaultClass *klass)
+{
+	GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
+	object_class->dispose = gtk_source_undo_manager_default_dispose;
+	object_class->finalize = gtk_source_undo_manager_default_finalize;
+
+	object_class->set_property = gtk_source_undo_manager_default_set_property;
+	object_class->get_property = gtk_source_undo_manager_default_get_property;
+
+	g_object_class_install_property (object_class,
+	                                 PROP_BUFFER,
+	                                 g_param_spec_object ("buffer",
+	                                                      "Buffer",
+	                                                      "The text buffer to add undo support on",
+	                                                      GTK_TYPE_TEXT_BUFFER,
+	                                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+
+	g_object_class_install_property (object_class,
+	                                 PROP_MAX_UNDO_LEVELS,
+	                                 g_param_spec_int ("max-undo-levels",
+	                                                   "Maximum Undo Levels",
+	                                                   "Number of undo levels for the buffer",
+	                                                   -1,
+	                                                   G_MAXINT,
+	                                                   DEFAULT_MAX_UNDO_LEVELS,
+	                                                   G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+}
+
+static void
+gtk_source_undo_manager_default_init (GtkSourceUndoManagerDefault *manager)
+{
+	manager->priv = gtk_source_undo_manager_default_get_instance_private (manager);
+	manager->priv->actions = g_ptr_array_new ();
+	manager->priv->next_redo = -1;
+}
+
 static void
 gtk_source_undo_manager_iface_init (GtkSourceUndoManagerIface *iface)
 {
