@@ -624,7 +624,7 @@ gtk_text_region_iterator_next (GtkTextRegionIterator *iter)
 	}
 }
 
-gboolean
+void
 gtk_text_region_iterator_get_subregion (GtkTextRegionIterator *iter,
 					GtkTextIter           *start,
 					GtkTextIter           *end)
@@ -632,31 +632,24 @@ gtk_text_region_iterator_get_subregion (GtkTextRegionIterator *iter,
 	GtkTextRegionIteratorReal *real;
 	Subregion *sr;
 
-	g_return_val_if_fail (iter != NULL, FALSE);
+	g_return_if_fail (iter != NULL);
 
 	real = (GtkTextRegionIteratorReal *)iter;
-	g_return_val_if_fail (check_iterator (real), FALSE);
-	g_return_val_if_fail (real->subregions != NULL, FALSE);
+	g_return_if_fail (check_iterator (real));
+	g_return_if_fail (real->subregions != NULL);
 
 	if (real->region->buffer == NULL)
 	{
-		return FALSE;
+		return;
 	}
 
 	sr = (Subregion*)real->subregions->data;
-	g_return_val_if_fail (sr != NULL, FALSE);
+	g_return_if_fail (sr != NULL);
 
-	if (start != NULL)
-	{
+	if (start)
 		gtk_text_buffer_get_iter_at_mark (real->region->buffer, start, sr->start);
-	}
-
-	if (end != NULL)
-	{
+	if (end)
 		gtk_text_buffer_get_iter_at_mark (real->region->buffer, end, sr->end);
-	}
-
-	return TRUE;
 }
 
 void
